import logging
import logging.config
import sys
from logging import StreamHandler
from logging.handlers import TimedRotatingFileHandler
from os.path import dirname, exists, expanduser, isdir, join, realpath

from aspen.configuration.optparser import validate_log_level


LOG_FORMAT = "%(message)s"
LOG_LEVEL = logging.WARNING
log = logging.getLogger('aspen.configuration.logging')


def configure_logging(opts, dotaspen, conf):
    """Given an opts object, configure logging.
    
    Logging can be configured from four places, in this order of 
    precedence:
    
      1) some other module (theoretically, I'm not sure an aspen user 
         could make this happen easily)
      2) the command line
      3) a logging.conf file
      4) a [logging] section in the aspen.conf file
    
    These are not layered; only one is used.

    """

    #FMT = "%(process)-6s%(levelname)-9s%(name)-14s%(message)s"

    kw = None
    logging_configured = bool(len(logging.getLogger().handlers))
                         # this test is taken from logging.basicConfig.

    if logging_configured:              # some other module
        log.warn("Logging is already configured, so aspen won't perform "
                 "any logging configuration.")

    if not logging_configured:          # command line
        kw = {'dotaspen': dotaspen}
        kw['filename'] = opts.log_file
        kw['filter'] = opts.log_filter
        kw['format'] = opts.log_format
        kw['level'] = opts.log_level
        if kw.values() != [None, None, None, None]: # at least one knob set
            if kw['format'] is None:
                kw['format'] = LOG_FORMAT
            if kw['level'] is None:
                kw['level'] = LOG_LEVEL
            simple_logging(**kw)
            log.info("Logging configured from the command line.")
            logging_configured = True

    if not logging_configured:          # logging.conf
        # TODO /etc/aspen/logging.conf
        # TODO /usr/local/etc/aspen/logging.conf
        logging_conf = join(dotaspen, 'logging.conf')
        if exists(logging_conf):
            logging.config.fileConfig(logging_conf) 
            log.info("logging configured from logging.conf")
            logging_configured = True

    if not logging_configured:          # aspen.conf [logging]
        kw = {'dotaspen': dotaspen}
        kw['filename'] = conf.logging.get('file')
        kw['filter'] = conf.logging.get('filter')
        kw['format'] = conf.logging.get('format', LOG_FORMAT)

        log_level = conf.logging.get('level')
        if log_level is not None:
            log_level = validate_log_level(log_level) 
        else:
            log_level = LOG_LEVEL
        kw['level'] = log_level

        simple_logging(**kw)
        log.info("Logging configured from aspen.conf.")
        logging_configured = True

    if kw is not None:
        return kw['filename'], kw['filter'], kw['format'], kw['level']
    else:
        return None, None, None, None


def simple_logging(dotaspen, filename, filter, format, level):
    """Used for configuring logging from the command line or aspen.conf.
    """
    handler = get_handler(dotaspen, filename)
    if filter is not None:
        filter = logging.Filter(filter)
        handler.addFilter(filter)
    formatter = logging.Formatter(fmt=format)
    handler.setFormatter(formatter)
    handler.setLevel(level)
    root_logger = logging.getLogger()
    root_logger.addHandler(handler)
    root_logger.setLevel(level) # bah

def get_handler(dotaspen, filename):
    # Handler
    # =======
    # sys.stdout or rotated file

    if filename is None:
        handler = StreamHandler(sys.stdout)
    else:
        # @@: Handle absolute paths on Windows
        #  http://sluggo.scrapping.cc/python/unipath/Unipath-current/unipath/abstractpath.py
        #  http://docs.python.org/library/os.path.html#os.path.splitunc
        if not filename.startswith('/'):
            filename = join(dotaspen, filename)
            filename = realpath(filename)
        logdir = dirname(filename)
        if not isdir(logdir):
            os.makedirs(logdir, 0644)
        handler = TimedRotatingFileHandler( filename=filename
                                          , when='midnight'
                                          , backupCount=7
                                           )
    return handler



def main(argv=None):
    """http://aspen.io/cli/
    """
    try:

        # Do imports.
        # ===========
        # These are in here so that if you Ctrl-C during an import, the
        # KeyboardInterrupt is caught and ignored. Yes, that's how much I care.
        # No, I don't care enough to put aspen/__init__.py in here too.

        import os
        import logging
        import socket
        import sys
        import time
        import traceback
        from os.path import exists, join

        from aspen import restarter
        from aspen.website import Website

        
        log = logging.getLogger('aspen.cli')


        # Website
        # =======
        # User-developers get this website object inside of their simplates and
        # hooks. It provides access to configuration information in addition to
        # being a WSGI callable and holding the request/response handling
        # logic. See aspen/website.py

        if argv is None:
            argv = sys.argv[1:]
        website = Website(argv)


        # Start serving the website.
        # ==========================
        # This amounts to binding the requested socket, with logging and 
        # restarting as needed. Wrap the whole thing in a try/except to
        # do some cleanup on shutdown.

        try:
            if hasattr(socket, 'AF_UNIX'):
                if website.sockfam == socket.AF_UNIX:
                    if exists(website.address):
                        log.info("Removing stale socket.")
                        os.remove(website.address)
            if website.port is not None:
                welcome = "port %d" % website.port
            else:
                welcome = website.address
            log.info("Starting %s engine." % website.engine.name)
            website.engine.bind()
            log.warn("Greetings, program! Welcome to %s." % welcome)
            if website.changes_kill:
                log.info("Aspen will die when files change.")
                restarter.install(website)
            website.start()
        except KeyboardInterrupt, SystemExit:
            pass
        except:
            traceback.print_exc()
        finally:
            if hasattr(socket, 'AF_UNIX'):
                if website.sockfam == socket.AF_UNIX:
                    if exists(website.address):
                        os.remove(website.address)
            website.stop()
    except KeyboardInterrupt, SystemExit:
        pass




import time

from aspen import Response
from aspen.sockets import TIMEOUT


class Transport(object):
    """A transport converts HTTP messages into Socket messages.
    """
    
    def __init__(self, socket):
        """Takes a Socket instance.
        """
        self.socket = socket


class XHRPollingTransport(Transport):

    state = 0
    timeout = TIMEOUT * 0.90 # Allow for some wiggle-room to prevent XHRs
                             # from cancelling too often.

    def respond(self, request):
        """Given a Request, return a Response.
        """
        request.allow('GET', 'POST')

        if self.state == 0:             # The client wants confirmation.
            response = Response(200, "1:::")
            self.state = 1

        elif request.method == 'POST':  # The client is sending us data.
            self.socket._send(request.body.raw)
            response = Response(200)
            
        elif request.method == 'GET':   # The client is asking for data.
            bytes_iter = iter([""])
            timeout = time.time() + self.timeout
            while time.time() < timeout:
                _bytes_iter = self.socket._recv()
                if _bytes_iter is not None:
                    bytes_iter = _bytes_iter
                    break
                request.engine.sleep(0.010)
            response = Response(200, bytes_iter)

        return response



import os
import sys
import time
from aspen.http.response import Response
from aspen import json_ as json


__version__ = "~~VERSION~~"
WINDOWS = sys.platform[:3] == 'win'
ENGINES = ['cherrypy', 'diesel', 'eventlet', 'pants', 'rocket']


def thrash():
    """This is a very simple tool to restart a process when it dies.

    It's designed to restart aspen in development when it dies because files
    have changed and you set changes_kill to 'yes' in the [aspen.cli] section
    of aspen.conf.

    http://aspen.io/thrash/

    """
    try:
        while 1:
            os.system(' '.join(sys.argv[1:]))
            time.sleep(1)
    except KeyboardInterrupt:
        pass
